use std::collections::HashMap;
use std::ffi::CStr;
use std::panic::catch_unwind;

use bstr::BStr;
use bstr::ByteSlice as _;
use color_eyre::Result;
use color_eyre::eyre;
use color_eyre::eyre::Context as _;

use crate::plugin::Identifier;
use crate::plugin::ModCallbacks;
use crate::plugin::Plugin;
use crate::stingray_sdk::LUA_REGISTRYINDEX;
use crate::stingray_sdk::{LuaState, LuaType, lua_State};
use crate::{LUA, global, plugin_mut};

pub const NAMESPACE_SEPARATOR: char = '.';

// As a minor optimization, only `catch_unwind` in debug mode.
#[cfg(debug_assertions)]
fn lua_wrapper<F>(l: *mut lua_State, f: F) -> i32
where
    F: FnOnce(&mut Plugin, &LuaState) -> Result<i32> + std::panic::UnwindSafe,
{
    let lua = global!(LUA);

    // We need to drop as many things as possible before the `lua_error` call
    // will longjmp and thereby ignoring Rust's cleanup.
    {
        let lua = LuaState::new(l, lua);

        let res = catch_unwind(|| {
            let mut plugin = plugin_mut!();
            f(&mut plugin, &lua)
        });
        match res {
            Ok(Ok(i)) => {
                return i;
            }
            Ok(Err(err)) => lua.pushstring(format!("{:?}", err)),
            Err(err) => lua.pushstring(format!("{:?}", err)),
        }
    }

    lua.error(l);
    0
}

#[cfg(not(debug_assertions))]
fn lua_wrapper<F>(l: *mut lua_State, f: F) -> i32
where
    F: FnOnce(&mut Plugin, &LuaState) -> Result<i32>,
{
    let lua = global!(LUA);

    // We need to drop as many things as possible before the `lua_error` call
    // will longjmp and thereby ignoring Rust's cleanup.
    {
        let lua = LuaState::new(l, lua);

        let mut plugin = plugin_mut!();
        match f(&mut plugin, &lua) {
            Ok(Ok(i)) => {
                return i;
            }
            Ok(Err(err)) => lua.pushstring(format!("{:?}", err)),
            Err(err) => lua.pushstring(format!("{:?}", err)),
        }
    }

    lua.error(l);
    0
}

/// Returns the namespace and name for a RPC.
///
/// The namespace may either be part of the name, separated by `NAMESPACE_SEPARATOR`
/// or omitted, in which case the mod name is used as fallback.
fn get_identifier<'a>(lua: &'a LuaState) -> Result<Identifier<'a>> {
    lua.getfield(1, c"name");
    if !lua.isstring(-1) {
        eyre::bail!("bad argument #1, not a mod object");
    }

    let mod_name = lua.tostring(-1);
    let name = lua.tostring(2);

    let (namespace, name) = name
        .split_once_str(&[NAMESPACE_SEPARATOR as u8])
        .map(|(namespace, name)| (BStr::new(namespace), BStr::new(name)))
        .unwrap_or((mod_name, name));

    Ok(Identifier::new(namespace, name))
}

// Called with `(mod: table, callbacks: table)`
pub extern "C" fn register_mod(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        let name = {
            lua.getfield(1, c"name");
            if !lua.isstring(-1) {
                eyre::bail!("bad argument #1, not a mod object");
            }

            lua.tostring(-1)
        };

        let callbacks = match lua.r#type(2) {
            LuaType::Nil => ModCallbacks::default(),
            LuaType::Table => {
                let get_callback = |name: &CStr| {
                    lua.getfield(2, name);

                    match lua.r#type(-1) {
                        LuaType::Nil => Ok(None),
                        LuaType::Function => Ok(Some(lua.lib_ref(LUA_REGISTRYINDEX))),
                        x => {
                            eyre::bail!(
                                "bad argument #2, expected a function for field {}, got {}",
                                name.to_string_lossy(),
                                x
                            );
                        }
                    }
                };

                ModCallbacks {
                    on_session_joined: get_callback(c"on_session_joined")?,
                    on_session_left: get_callback(c"on_session_left")?,
                    on_user_joined: get_callback(c"on_user_joined")?,
                    on_user_left: get_callback(c"on_user_left")?,
                }
            }
            x => {
                eyre::bail!("bad argument #2, expected a table got {}", x);
            }
        };

        plugin.register_mod(name, callbacks);

        // TODO: Register unload and disable handlers?
        // Will likely have to monkey-patch the mod object.
        // DMF doesn't expose a good way to hook into it, and it wouldn't be feasible to expect a
        // DMF update to allow it.

        Ok(0)
    })
}

// Called with `(mod: table, name: string, opts: table)`
pub extern "C" fn create_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        lua.getfield(1, c"version");
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.istable(3) {
            eyre::bail!("bad argument #3, expected a table");
        }

        let _version = match lua.r#type(-1) {
            LuaType::Nil => {
                eyre::bail!("bad argument #1, mod object doesn't provide a 'version' field");
            }
            LuaType::Number => lua.tostring(-1),
            LuaType::String => {
                let version = lua.tostring(-1);
                // TODO: parse as date or semver-like
                version
            }
            x => {
                eyre::bail!(
                    "invalid type {} for 'version' field, must be number or a version string",
                    x
                )
            }
        };

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;
        plugin.create_rpc(id);

        Ok(0)
    })
}

type LuaMap<'a> = HashMap<LuaKey<'a>, LuaValue<'a>>;

#[derive(bincode::BorrowDecode, bincode::Encode, PartialEq)]
pub enum LuaKey<'a> {
    String(&'a [u8]),
    Number(f64),
}

impl Eq for LuaKey<'_> {}

impl std::hash::Hash for LuaKey<'_> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        match self {
            LuaKey::String(s) => s.hash(state),
            LuaKey::Number(n) => (*n as u128).hash(state),
        }
    }
}

#[derive(bincode::BorrowDecode, bincode::Encode)]
pub enum LuaValue<'a> {
    Nil,
    Boolean(bool),
    Number(f64),
    String(&'a [u8]),
    Table(LuaMap<'a>),
}

impl<'a> LuaValue<'a> {
    fn parse_table(lua: &'a LuaState, idx: i32) -> Result<Self> {
        let mut values = LuaMap::default();

        // `nil` as initial key tells `lua_next` to pick the first key.
        lua.pushnil();

        loop {
            if lua.next(idx) <= 0 {
                break;
            }

            let key = match lua.r#type(-2) {
                LuaType::Number => LuaKey::Number(lua.tonumber(-2)),
                LuaType::String => LuaKey::String(lua.tostring(-2)),
                t => eyre::bail!("Unsupported type {} for args table key", t),
            };

            let value = Self::get_from_stack(lua, -1)?;

            values.insert(key, value);

            // Pop the value, but keep the key so `lua_next` knows where to continue
            lua.pop();
        }

        // Pop the last key
        lua.pop();

        Ok(LuaValue::Table(values))
    }

    /// Gets the value from the stack at `idx`
    pub fn get_from_stack(lua: &'a LuaState, idx: i32) -> Result<Self> {
        match lua.r#type(idx) {
            LuaType::None | LuaType::Nil => Ok(LuaValue::Nil),
            LuaType::Boolean => Ok(LuaValue::Boolean(lua.toboolean(-1))),
            LuaType::Number => Ok(LuaValue::Number(lua.tonumber(-1))),
            LuaType::String => Ok(LuaValue::String(lua.tostring(-1))),
            LuaType::Table => Self::parse_table(lua, -1),
            x => eyre::bail!("Unknown or unsupported Lua type {}", x),
        }
    }

    /// Pushes the value onto a Lua stack
    pub fn push_to_stack(&self, lua: &LuaState) {
        match self {
            LuaValue::Nil => lua.pushnil(),
            LuaValue::Boolean(bool) => lua.pushboolean(*bool),
            LuaValue::Number(num) => lua.pushnumber(*num),
            LuaValue::String(s) => lua.pushstring(*s),
            LuaValue::Table(table) => Self::push_table(lua, table),
        }
    }

    fn push_table(lua: &LuaState, table: &LuaMap) {
        let (narr, nrec) = table.iter().fold((0, 0), |(narr, nrec), (k, _)| match k {
            LuaKey::String(_) => (narr, nrec + 1),
            LuaKey::Number(_) => (narr + 1, nrec),
        });

        lua.createtable(narr, nrec);
        let tbl_index = lua.gettop();

        for (k, v) in table.iter() {
            match k {
                LuaKey::String(s) => lua.pushstring(*s),
                LuaKey::Number(num) => lua.pushnumber(*num),
            }

            v.push_to_stack(lua);

            lua.settable(tbl_index);
        }
    }
}

// Called with `(mod: table, name: string, args: table)`
pub extern "C" fn send_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.istable(3) {
            eyre::bail!("bad argument #3, expected a table");
        }

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;
        let args = LuaValue::parse_table(lua, 3).wrap_err("Failed to parse args table")?;

        plugin.send_rpc(id, args).wrap_err("Failed to send RPC")?;

        Ok(0)
    })
}

// Called with `(mod: table, name: string, callback: Fn(args))`
pub extern "C" fn receive_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.isfunction(3) {
            eyre::bail!("bad argument #3, expected a function");
        }

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;

        // We will utilize Lua's _references_ to store the callback functions. Since we're doing
        // the bookkeeping on the Rust side, we don't need to worry about any structured layout on
        // the Lua side.

        lua.pushvalue(3);
        let fn_ref = lua.lib_ref(LUA_REGISTRYINDEX);

        plugin
            .add_rpc_listener(id, fn_ref)
            .wrap_err("Failed to add RPC listener")?;

        Ok(0)
    })
}

// Called with `(server_name: string, peer_id: string, player_id: string)`
pub extern "C" fn join_session(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.isstring(1) {
            eyre::bail!("bad argument #1, expected a string");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.isstring(3) {
            eyre::bail!("bad argument #3, expected a string");
        }

        let server_name = lua.tostring(1);
        let peer_id = lua.tostring(2);
        let player_id = lua.tostring(3);

        plugin
            .join_session(lua, server_name, peer_id, player_id)
            .wrap_err("Failed to join session")?;

        Ok(0)
    })
}

// Called with `()`
pub extern "C" fn leave_session(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        plugin
            .leave_session(lua)
            .wrap_err("Failed to leave session")?;

        Ok(0)
    })
}
