use std::collections::HashMap;
use std::hash::{DefaultHasher, Hash, Hasher as _};
use std::mem;
use std::ops::Deref;
use std::thread;
use std::thread::JoinHandle;

use bstr::{BStr, BString};
use color_eyre::Result;
use color_eyre::eyre;
use color_eyre::eyre::Context as _;
use libp2p::PeerId;
use libp2p::futures::StreamExt;
use libp2p::gossipsub;
use libp2p::gossipsub::{
    AllowAllSubscriptionFilter, Event, IdentTopic, IdentityTransform, Message, TopicHash,
};
use libp2p::noise;
use libp2p::swarm::SwarmEvent;
use libp2p::tcp;
use libp2p::yamux;
use tokio::sync::mpsc;
use tokio::sync::mpsc::Receiver;
use tokio::sync::mpsc::UnboundedSender;
use tokio::sync::mpsc::error::TryRecvError;

use crate::lua;
use crate::lua::{LuaValue, NAMESPACE_SEPARATOR};
use crate::rpc::RPC;
use crate::stingray_sdk::{LUA_REGISTRYINDEX, LuaRef, LuaState};
use crate::{CHANNEL_BUFFER_SIZE, EVENTS_PER_TICK};
use crate::{LUA, MODULE_NAME, global};

type ModName = BString;

#[derive(Clone, Copy, Debug)]
pub struct Identifier<'a> {
    name: &'a BStr,
    namespace: &'a BStr,
}

impl<'a> Identifier<'a> {
    pub fn new(namespace: &'a BStr, name: &'a BStr) -> Self {
        Self { namespace, name }
    }
}

impl From<Identifier<'_>> for String {
    fn from(value: Identifier<'_>) -> Self {
        format!("{}", value)
    }
}

impl std::fmt::Display for Identifier<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}{}{}", self.namespace, NAMESPACE_SEPARATOR, self.name)
    }
}

#[derive(Debug, bincode::Encode, bincode::Decode)]
// To avoid having to implement `bincode` for `bstr` types, we use raw `Vec<u8>` here,
// which can be turned into `BString` cheaply.
pub(crate) enum SwarmMessage {
    // To avoid extra cloning, or having to deal with lifetimes,
    // this message type contains the Lua args already encoded.
    Rpc(Vec<u8>),
    SessionJoin {
        /// The in-game player ID for this peer
        player_id: Vec<u8>,
        /// A list of mod names this peer has enabled
        mods: Vec<Vec<u8>>,
    },
    SessionLeave,
}

#[derive(Debug)]
pub(crate) enum SwarmTask {
    Message { topic: Topic, msg: SwarmMessage },
    Subscribe(Topic),
    Unsubscribe(Vec<Topic>),
}

#[derive(Clone, Debug)]
pub(crate) struct Topic(IdentTopic);

impl Eq for Topic {}
impl PartialEq for Topic {
    fn eq(&self, other: &Self) -> bool {
        self.0.hash() == other.0.hash()
    }
}

impl Hash for Topic {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        self.0.hash().hash(state);
    }
}

impl std::fmt::Display for Topic {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.0)
    }
}

impl std::ops::Deref for Topic {
    type Target = IdentTopic;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

impl From<Topic> for TopicHash {
    fn from(value: Topic) -> Self {
        Self::from(value.0)
    }
}

#[derive(Debug)]
struct Player {
    id: BString,
    mods: Vec<ModName>,
}

#[derive(Default)]
pub(crate) struct ModCallbacks {
    pub on_session_joined: Option<LuaRef>,
    pub on_session_left: Option<LuaRef>,
    pub on_user_joined: Option<LuaRef>,
    pub on_user_left: Option<LuaRef>,
}

struct Session {
    online: bool,
    address: String,
    peer_id: BString,
    topics: HashMap<TopicHash, Topic>,
    peers: HashMap<PeerId, Player>,
}

impl Session {
    pub fn new_topic(&mut self, topic: impl Into<String>) -> &Topic {
        self.topics
            .entry(TopicHash::from_raw(topic.into()))
            .or_insert_with_key(|topic| {
                Topic(IdentTopic::new(format!("{}/{}", self.address, topic)))
            })
    }

    pub fn get_topic(&self, topic: impl Into<String>) -> Option<&Topic> {
        let hash = TopicHash::from_raw(topic.into());
        self.topics.get(&hash)
    }
}

pub(crate) struct Plugin {
    swarm_thread: JoinHandle<Result<()>>,
    event_rx: Receiver<SwarmEvent<Event>>,
    send_tx: UnboundedSender<SwarmTask>,
    session: Session,
    rpcs: HashMap<Topic, RPC>,
    mods: HashMap<ModName, ModCallbacks>,
}

impl Plugin {
    pub fn new() -> Result<Self> {
        let mut swarm = libp2p::SwarmBuilder::with_new_identity()
            .with_tokio()
            .with_tcp(
                tcp::Config::default(),
                noise::Config::new,
                yamux::Config::default,
            )
            .wrap_err("Failed to configure transport")?
            .with_behaviour(|key| {
                let message_id_fn = |msg: &gossipsub::Message| {
                    let mut s = DefaultHasher::new();
                    msg.data.hash(&mut s);
                    gossipsub::MessageId::from(s.finish().to_string())
                };

                let config = gossipsub::ConfigBuilder::default()
                    .validation_mode(gossipsub::ValidationMode::Strict)
                    .message_id_fn(message_id_fn)
                    .build()
                    .wrap_err("Failed to create gossipsub config")?;

                // TODO: Benchmark if compression would be beneficial.
                let behavior: gossipsub::Behaviour<IdentityTransform, AllowAllSubscriptionFilter> =
                    gossipsub::Behaviour::new(
                        gossipsub::MessageAuthenticity::Signed(key.clone()),
                        config,
                    )?;

                Ok(behavior)
            })
            .wrap_err("Failed to configure behavior")?
            .build();

        swarm
            .listen_on(
                "/ipv4/0.0.0.0/tcp/0"
                    .parse()
                    .expect("listen address invalid"),
            )
            .wrap_err("failed to listen on IPv4 socket")?;

        // TODO: Connect to first hard coded relay server(s), which should then
        // fan out to online peers.
        // Maybe a cheap server (e.g. AWS free tier instance) is already enough to handle all
        // session negotiations. If not, then a system is needed where each mod user also becomes
        // available for that, and the hosted server merely acts as load balancer between them.
        //
        // swarm.dial(RELAY_SERVER)

        let (event_tx, event_rx) = mpsc::channel::<SwarmEvent<Event>>(CHANNEL_BUFFER_SIZE);
        // Since the `Sender` will be on the sync side, and we want to avoid blocking, we make this
        // unbounded to limit the chance for it to actually block.
        let (send_tx, send_rx) = mpsc::unbounded_channel::<SwarmTask>();

        let swarm_thread = thread::Builder::new()
            .name("p2p-swarm".into())
            .spawn(move || {
                let rt = tokio::runtime::Builder::new_current_thread()
                    .build()
                    .wrap_err("Failed to create tokio runtime")?;

                let mut swarm = swarm;
                let mut send_rx = send_rx;

                rt.block_on(async move {
                    loop {
                        tokio::select! {
                            event = swarm.select_next_some() => {
                                event_tx.send(event).await.wrap_err("Failed to queue event")?;
                            }
                            task = send_rx.recv() => {
                                let Some(task) = task else {
                                    eyre::bail!("Task queue closed prematurely");
                                };

                                let behavior = swarm.behaviour_mut();

                                match task{
                                    SwarmTask::Message { topic, msg } => {
                                        let data = bincode::encode_to_vec(msg, bincode::config::standard())
                                            .wrap_err("Failed to encode swarm message")?;
                                        behavior
                                            .publish(topic, data)
                                            .wrap_err("Failed to send message")?;
                                    },
                                    SwarmTask::Subscribe(topic) => {
                                        behavior
                                            .subscribe(&topic)
                                            .wrap_err_with(|| format!("Failed to subscribe to topic {}", topic))?;
                                    },
                                    SwarmTask::Unsubscribe(topics) => {
                                        for topic in topics {
                                            behavior.unsubscribe(&topic);
                                        }
                                    }
                                }
                            }
                        }
                    }
                })
            })
            .wrap_err("Failed to create p2p-swarm thread")?;

        log::info!("Plugin initialized");

        Ok(Self {
            swarm_thread,
            event_rx,
            send_tx,
            session: Session {
                online: false,
                address: Default::default(),
                peer_id: Default::default(),
                topics: Default::default(),
                peers: Default::default(),
            },
            rpcs: Default::default(),
            mods: Default::default(),
        })
    }

    pub fn setup_game(&self) {
        let lua = global!(LUA);
        // Internal API
        // TODO: Maybe useful to move these to a separate module, e.g. `P2PInternal`?
        lua.add_module_function(MODULE_NAME, c"join_session", lua::join_session);
        lua.add_module_function(MODULE_NAME, c"leave_session", lua::leave_session);

        // User-facing API
        lua.add_module_function(MODULE_NAME, c"register_mod", lua::register_mod);
        lua.add_module_function(MODULE_NAME, c"create_rpc", lua::create_rpc);
        lua.add_module_function(MODULE_NAME, c"send_rpc", lua::send_rpc);
        lua.add_module_function(MODULE_NAME, c"receive_rpc", lua::receive_rpc);

        log::debug!("Lua functions registered");
    }

    pub fn shutdown_game(&self) {
        // TODO: Find a way to send a close command, and make it blocking.
    }

    pub fn update_game(&mut self, dt: f32) {
        if let Err(err) = self.update(dt) {
            log::error!("{:?}", err);
            // TODO: Exit application or find a better way to report the error
        }
    }

    pub fn update(&mut self, _dt: f32) -> Result<()> {
        if self.swarm_thread.is_finished() {
            eyre::bail!("p2p-swarm thread terminated prematurely",);
            // TODO: Move `self.swarm_thread` into a data structure that I can move it out of here,
            // so that I can call `.join` and get the error it ended with.
        }

        let lua = {
            let lua = global!(LUA);
            let l = lua.getscriptenvironmentstate();
            LuaState::new(l, lua)
        };

        // Limit the amount of events we handle per tick to avoid potential large spikes in
        // frame time.
        // TODO: Add a system to monitor the amount of incoming vs processed events, and add
        // warnings when this cannot catch up.
        // TODO: Check if it might become necessary to create a more elaborate system that can
        // change this limit dynamically based on the amount of incoming messages.
        // This must be able to eventually catch up.
        for _ in 0..EVENTS_PER_TICK {
            let event = match self.event_rx.try_recv() {
                Ok(event) => event,
                Err(TryRecvError::Empty) => break,
                Err(TryRecvError::Disconnected) => {
                    eyre::bail!("p2p-swarm channel disconnected prematurely");
                }
            };

            match event {
                SwarmEvent::Behaviour(Event::Message {
                    propagation_source,
                    message_id,
                    message:
                        Message {
                            topic,
                            data,
                            source,
                            sequence_number: _,
                        },
                }) => {
                    log::debug!(
                        "Received message {message_id} from {propagation_source} for topic {}",
                        topic
                    );

                    let Some(topic) = self.session.topics.get(&topic) else {
                        log::warn!("Received message for unknown topic {}", topic);
                        continue;
                    };

                    let Ok((msg, _)) = bincode::borrow_decode_from_slice::<SwarmMessage, _>(
                        &data,
                        bincode::config::standard(),
                    ) else {
                        log::error!(
                            "Failed to decode data for message {message_id} on topic {}",
                            topic
                        );
                        continue;
                    };

                    match msg {
                        SwarmMessage::Rpc(rpc_args) => {
                            let Some(rpc) = self.rpcs.get(topic) else {
                                log::warn!("Topic {} does not have an RPC", topic);
                                continue;
                            };

                            let Ok((args, _)) = bincode::borrow_decode_from_slice::<LuaValue<'_>, _>(
                                &rpc_args,
                                bincode::config::standard(),
                            ) else {
                                log::error!(
                                    "Failed to decode data for message {message_id} on topic {}",
                                    topic
                                );
                                continue;
                            };

                            args.push_to_stack(&lua);
                            let args_index = lua.gettop();

                            for fn_ref in rpc.listeners() {
                                lua.rawgeti(LUA_REGISTRYINDEX, *fn_ref);
                                lua.pushvalue(args_index);
                                lua.pcall(1, 0).wrap_err_with(|| {
                                    format!("Failed to call listener for RPC '{}'", topic)
                                })?;
                            }
                        }
                        SwarmMessage::SessionJoin { player_id, mods } => match source {
                            Some(peer_id) => {
                                let id = BString::new(player_id);
                                let mods = mods.into_iter().map(ModName::new).collect::<Vec<_>>();
                                let peer = Player { id, mods };

                                // TODO: Check if a corresponding player is in the lobby

                                {
                                    let callbacks = self
                                        .mods
                                        .iter()
                                        .filter(|(name, _)| peer.mods.contains(name))
                                        .filter_map(|(_, cbs)| cbs.on_user_joined);

                                    lua.pushstring(peer.id.clone());
                                    let arg_index = lua.gettop();

                                    for fn_ref in callbacks {
                                        lua.pushnumber(fn_ref as f64);
                                        lua.gettable(LUA_REGISTRYINDEX);
                                        lua.pushvalue(arg_index);
                                        lua.pcall(1, 0)
                                            .wrap_err("on_user_joined handler failed")?;
                                    }
                                }

                                self.session.peers.insert(peer_id, peer);
                            }
                            None => {
                                log::warn!(
                                    "Got SessionJoin event without a source peer_id. Don't know how to handle."
                                );
                            }
                        },
                        SwarmMessage::SessionLeave => match source {
                            Some(peer_id) => {
                                if let Some(peer) = self.session.peers.remove(&peer_id) {
                                    let callbacks = self
                                        .mods
                                        .iter()
                                        .filter(|(name, _)| peer.mods.contains(name))
                                        .filter_map(|(_, cbs)| cbs.on_user_left);

                                    lua.pushstring(peer.id);
                                    let arg_index = lua.gettop();

                                    for fn_ref in callbacks {
                                        lua.pushnumber(fn_ref as f64);
                                        lua.gettable(LUA_REGISTRYINDEX);
                                        lua.pushvalue(arg_index);
                                        lua.pcall(1, 0).wrap_err("on_user_left handler failed")?;
                                    }
                                }
                            }
                            None => {
                                log::warn!(
                                    "Got SessionLeave event without a source peer_id. Don't know how to handle."
                                );
                            }
                        },
                    }
                }
                SwarmEvent::Behaviour(Event::GossipsubNotSupported { peer_id }) => {
                    log::warn!(
                        "Peer {peer_id} does not support Gossipsub. Are they really a dt-p2p user?"
                    );
                }
                SwarmEvent::ConnectionEstablished {
                    peer_id,
                    connection_id,
                    endpoint: _,
                    num_established,
                    concurrent_dial_errors: _,
                    established_in,
                } => {
                    log::info!(
                        "Connection {} established with {} in {} ms. Total: {}",
                        connection_id,
                        peer_id,
                        established_in.as_millis(),
                        num_established
                    );
                    // TODO: Start a timeout to wait for a `SwarmMessage::SessionJoin` to show a
                    // warning when a client connected but didn't join the session.
                    // Not sure if that should ever happen.
                }
                SwarmEvent::ConnectionClosed {
                    peer_id,
                    connection_id,
                    endpoint: _,
                    num_established,
                    cause,
                } => {
                    log::info!(
                        "Connection {} with {} closed for reason {:?}. Total: {}",
                        connection_id,
                        peer_id,
                        cause,
                        num_established
                    );
                    if self.session.peers.contains_key(&peer_id) {
                        log::warn!("Peer dropped connection without properly leaving session!");
                        // TODO: Start a timeout and if the peer doesn't come back, remove the peer
                        // and trigger "user left" callbacks.
                        // TODO: Maybe also check if a corresponding player is still in the game
                        // lobby.
                    }
                }
                SwarmEvent::IncomingConnectionError {
                    connection_id,
                    local_addr: _,
                    send_back_addr: _,
                    error,
                } => {
                    log::error!("Error on incoming connection {connection_id}: {error}");
                }
                SwarmEvent::OutgoingConnectionError {
                    connection_id,
                    peer_id: _,
                    error,
                } => {
                    log::error!("Error on outgoing connection {connection_id}: {error}");
                }
                SwarmEvent::ListenerError { listener_id, error } => {
                    log::error!("Listener {listener_id} failed: {error}");
                }
                SwarmEvent::NewListenAddr {
                    listener_id,
                    address,
                } => {
                    log::debug!("Listening on {address} with ID {listener_id}");
                }
                // TODO: Maybe add tracing information for the events we don't handle
                _ => {}
            }
        }

        Ok(())
    }

    pub(crate) fn register_mod(&mut self, name: &BStr, callbacks: ModCallbacks) {
        self.mods.insert(name.to_owned(), callbacks);
    }

    pub(crate) fn create_rpc(&mut self, id: Identifier) {
        let rpc = RPC::new();
        let topic = self.session.new_topic(id);
        self.rpcs.insert(topic.clone(), rpc);
    }

    pub(crate) fn send_rpc(&self, id: Identifier, args: LuaValue<'_>) -> Result<()> {
        let Some(topic) = self.session.get_topic(id) else {
            eyre::bail!("Topic for this RPC does not exist");
        };

        if !self.rpcs.contains_key(topic) {
            eyre::bail!("RPC '{}' does not exist", id);
        }

        let args = bincode::encode_to_vec(args, bincode::config::standard())
            .wrap_err("Failed to encode RPC args")?;

        // TODO: Add metrics/profiling for size of the encoded args

        let msg = SwarmTask::Message {
            topic: topic.clone(),
            msg: SwarmMessage::Rpc(args),
        };
        self.send_tx
            .send(msg)
            .wrap_err("Failed to queue RPC call")?;

        Ok(())
    }

    pub(crate) fn add_rpc_listener(&mut self, id: Identifier, fn_ref: LuaRef) -> Result<()> {
        let topic = self.session.new_topic(id);
        let Some(rpc) = self.rpcs.get_mut(topic) else {
            eyre::bail!("RPC '{}' does not exist", id);
        };

        rpc.add_listener(fn_ref);
        self.send_tx
            .send(SwarmTask::Subscribe(topic.clone()))
            .wrap_err("Failed to subscribe to RPC topic")?;

        Ok(())
    }

    pub(crate) fn join_session(
        &mut self,
        lua: &LuaState,
        server_name: &BStr,
        peer_id: &BStr,
        player_id: &BStr,
    ) -> Result<()> {
        self.session.peer_id = peer_id.to_owned();
        self.session.address = format!("/{:08X}", hash(server_name));

        let session_topic = self.session.new_topic("session").clone();

        self.send_tx
            .send(SwarmTask::Subscribe(session_topic.clone()))
            .wrap_err("Failed to subscribe to session topic")?;

        // TODO: Is there a way to do this without the intermediate `Vec`?
        // Probably not, since we cannot keep an immutable reference on `self.rpcs`, while
        // also trying to get a mutable reference on `self` for `self.subscribe`.
        let topics = self.rpcs.keys().cloned().collect::<Vec<_>>();
        for topic in topics {
            self.send_tx
                .send(SwarmTask::Subscribe(topic.clone()))
                .wrap_err_with(|| format!("Failed to subscribe to {}", topic))?;
        }

        self.session.online = true;

        let msg = SwarmTask::Message {
            topic: session_topic,
            msg: SwarmMessage::SessionJoin {
                player_id: player_id.to_vec(),
                mods: self.mods.keys().map(|name| name.to_vec()).collect(),
            },
        };
        self.send_tx
            .send(msg)
            .wrap_err("Failed to queue RPC call")?;

        // TODO: Do I want to wait for a reply or some sort of confirmation?
        // Currently, we just assume this works always.

        for fn_ref in self.mods.values().filter_map(|cbs| cbs.on_session_joined) {
            lua.pushnumber(fn_ref as f64);
            lua.gettable(LUA_REGISTRYINDEX);
            lua.pcall(0, 0)
                .wrap_err("on_session_joined handler failed")?;
        }

        Ok(())
    }

    pub(crate) fn leave_session(&mut self, lua: &LuaState) -> Result<()> {
        if !self.session.online {
            log::warn!("There is no session to leave");
            return Ok(());
        }

        let topics = mem::take(&mut self.session.topics);
        self.send_tx
            .send(SwarmTask::Unsubscribe(topics.into_values().collect()))
            .wrap_err("Failed to queue unsubscribe task")?;

        self.session.online = false;

        for fn_ref in self.mods.values().filter_map(|cbs| cbs.on_session_left) {
            lua.pushnumber(fn_ref as f64);
            lua.gettable(LUA_REGISTRYINDEX);
            lua.pcall(0, 0).wrap_err("on_session_left handler failed")?;
        }

        Ok(())
    }
}

impl std::fmt::Debug for Plugin {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("PluginApi")
    }
}

fn hash(val: impl AsRef<[u8]>) -> u64 {
    let mut hasher = DefaultHasher::new();
    hasher.write(val.as_ref());
    hasher.finish()
}
